//*****************************************************************
/*
  QJackTrip: Bringing a graphical user interface to JackTrip, a
  system for high quality audio network performance over the
  internet.

  Copyright (c) 2020 Aaron Wyatt.

  This file is part of QJackTrip.

  QJackTrip is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  QJackTrip is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with QJackTrip.  If not, see <https://www.gnu.org/licenses/>.
*/
//*****************************************************************

#ifndef ABOUT_H
#define ABOUT_H

#include <QDialog>
#include <QScopedPointer>
#include <QTextDocument>

namespace Ui
{
class About;
}

class About : public QDialog
{
    Q_OBJECT

   public:
    explicit About(QWidget* parent = nullptr);
    ~About() override;

    void resizeEvent(QResizeEvent* event) override;

    static const QString s_buildType;
    static const QString s_buildID;

   private:
    QScopedPointer<Ui::About> m_ui;
    QTextDocument aboutText;
};

#endif  // ABOUT_H
